#![feature(concat_idents)]

mod camera_model;
mod cava_rust;
mod image_proc;

use self::camera_model::*;
use self::cava_rust::CHAN;
use self::image_proc::*;

use hercules_rt::{runner, HerculesImmBox, HerculesImmBoxTo, HerculesMutBox};

use image::ImageError;

use clap::Parser;

juno_build::juno!("cava");

// Individual lifetimes are not needed in this example but should probably be generated for
// flexibility
async fn safe_run<'a, 'b: 'a, 'c: 'a, 'd: 'a, 'e: 'a, 'f: 'a, 'g: 'a>(
    runner: &'a mut HerculesRunner_cava,
    r: u64,
    c: u64,
    num_ctrl_pts: u64,
    input: &'b HerculesImmBox<'b, u8>,
    tstw: &'c HerculesImmBox<'c, f32>,
    ctrl_pts: &'d HerculesImmBox<'d, f32>,
    weights: &'e HerculesImmBox<'e, f32>,
    coefs: &'f HerculesImmBox<'f, f32>,
    tonemap: &'g HerculesImmBox<'g, f32>,
) -> HerculesMutBox<'a, u8> {
    HerculesMutBox::from(
        runner
            .run(
                r,
                c,
                num_ctrl_pts,
                input.to(),
                tstw.to(),
                ctrl_pts.to(),
                weights.to(),
                coefs.to(),
                tonemap.to(),
            )
            .await,
    )
}

fn run_cava(
    rows: usize,
    cols: usize,
    num_ctrl_pts: usize,
    image: &[u8],
    tstw: &[f32],
    ctrl_pts: &[f32],
    weights: &[f32],
    coefs: &[f32],
    tonemap: &[f32],
) -> Box<[u8]> {
    assert_eq!(image.len(), CHAN * rows * cols);
    assert_eq!(tstw.len(), CHAN * CHAN);
    assert_eq!(ctrl_pts.len(), num_ctrl_pts * CHAN);
    assert_eq!(weights.len(), num_ctrl_pts * CHAN);
    assert_eq!(coefs.len(), 4 * CHAN);
    assert_eq!(tonemap.len(), 256 * CHAN);

    let image = HerculesImmBox::from(image);
    let tstw = HerculesImmBox::from(tstw);
    let ctrl_pts = HerculesImmBox::from(ctrl_pts);
    let weights = HerculesImmBox::from(weights);
    let coefs = HerculesImmBox::from(coefs);
    let tonemap = HerculesImmBox::from(tonemap);

    let mut r = runner!(cava);

    async_std::task::block_on(async {
        safe_run(
            &mut r,
            rows as u64,
            cols as u64,
            num_ctrl_pts as u64,
            &image,
            &tstw,
            &ctrl_pts,
            &weights,
            &coefs,
            &tonemap,
        )
        .await
    })
    .as_slice()
    .to_vec()
    .into_boxed_slice()
}

enum Error {
    IOError(std::io::Error),
    ImageError(image::ImageError),
    ParseIntError(std::num::ParseIntError),
    ParseFloatError(std::num::ParseFloatError),
}

impl std::fmt::Debug for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Error::IOError(err) => write!(f, "IO Error: {:?}", err),
            Error::ImageError(err) => write!(f, "Image Error: {:?}", err),
            Error::ParseIntError(err) => write!(f, "Parse Error: {:?}", err),
            Error::ParseFloatError(err) => write!(f, "Parse Error: {:?}", err),
        }
    }
}

impl From<std::io::Error> for Error {
    fn from(value: std::io::Error) -> Self {
        Error::IOError(value)
    }
}

impl From<ImageError> for Error {
    fn from(value: ImageError) -> Self {
        Error::ImageError(value)
    }
}

impl From<std::num::ParseIntError> for Error {
    fn from(value: std::num::ParseIntError) -> Self {
        Error::ParseIntError(value)
    }
}

impl From<std::num::ParseFloatError> for Error {
    fn from(value: std::num::ParseFloatError) -> Self {
        Error::ParseFloatError(value)
    }
}

#[derive(Parser)]
#[clap(author, version, about, long_about = None)]
struct CavaInputs {
    input: String,
    #[clap(short, long, value_name = "PATH")]
    output: Option<String>,
    #[clap(short, long)]
    verify: bool,
    #[clap(long = "output-verify", value_name = "PATH")]
    output_verify: Option<String>,
    cam_model: String,
    #[clap(short, long)]
    crop_rows: Option<usize>,
    #[clap(short, long)]
    crop_cols: Option<usize>,
}

fn cava_harness(args: CavaInputs) {
    let CavaInputs {
        input,
        output,
        verify,
        output_verify,
        cam_model,
        crop_rows,
        crop_cols,
    } = args;
    let RawImage { rows, cols, pixels } =
        read_raw(input, crop_rows, crop_cols).expect("Error loading image");
    let CamModel {
        tstw,
        num_ctrl_pts,
        ctrl_pts,
        weights,
        coefs,
        tonemap,
    } = load_cam_model(cam_model, CHAN).expect("Error loading camera model");

    println!(
        "Running cava with {} rows, {} columns, and {} control points.",
        rows, cols, num_ctrl_pts
    );
    let result = run_cava(
        rows,
        cols,
        num_ctrl_pts,
        &pixels,
        &tstw,
        &ctrl_pts,
        &weights,
        &coefs,
        &tonemap,
    );

    if let Some(output) = output {
        extern_image(rows, cols, &*result)
            .save(output)
            .expect("Error saving image");
    }

    if verify {
        let cpu_result = cava_rust::cava(
            rows,
            cols,
            num_ctrl_pts,
            &pixels,
            &tstw,
            &ctrl_pts,
            &weights,
            &coefs,
            &tonemap,
        );

        if let Some(output) = output_verify {
            extern_image(rows, cols, &cpu_result)
                .save(output)
                .expect("Error saving verification image");
        }

        let max_diff = result
            .iter()
            .zip(cpu_result.iter())
            .map(|(a, b)| (*a as i16 - *b as i16).abs())
            .max()
            .unwrap_or(0);

        assert!(
            max_diff <= 3,
            "Verification failed: maximum pixel difference of {} exceeds threshold of 3",
            max_diff
        );
    }
}

fn main() {
    let args = CavaInputs::parse();
    cava_harness(args);
}

#[test]
fn cava_test_small() {
    cava_harness(CavaInputs {
        input: "examples/raw_tulip-small.bin".to_string(),
        output: None,
        verify: true,
        output_verify: None,
        cam_model: "cam_models/NikonD7000".to_string(),
        crop_rows: Some(144),
        crop_cols: Some(192),
    });
}

#[test]
#[ignore]
fn cava_test_full() {
    cava_harness(CavaInputs {
        input: "examples/raw_tulips.bin".to_string(),
        output: None,
        verify: true,
        output_verify: None,
        cam_model: "cam_models/NikonD7000".to_string(),
        crop_rows: None,
        crop_cols: None,
    });
}
